import inspect
import typing

import pytest
from pydal.objects import Query

from src.typedal import TypeDAL, TypedField, TypedTable, relationship
from typedal.types import CacheFn, CacheModel, CacheTuple, Rows

db = TypeDAL("sqlite:memory")


@db.define()
class TestQueryTable(TypedTable):
    number: TypedField[int]
    other = TypedField(str, default="Something")
    yet_another = TypedField(list[str], default=["something", "and", "other", "things"])

    relations = relationship(
        list["TestRelationship"], condition=lambda self, other: self.id == other.querytable, join="left"
    )


@db.define()
class TestRelationship(TypedTable):
    name: TypedField[str]
    value: TypedField[int]

    querytable: TestQueryTable


class Undefined(TypedTable):
    value: int


def test_repr_unbound():
    assert "unbound table Undefined" in str(Undefined)


def test_query_type():
    assert isinstance(TestQueryTable.number > 3, Query)
    assert isinstance(TestQueryTable.number >= 3, Query)
    assert isinstance(TestQueryTable.number == 3, Query)
    assert isinstance(TestQueryTable.number < 3, Query)
    assert isinstance(TestQueryTable.number <= 3, Query)
    assert isinstance(TestQueryTable.number != 3, Query)


"""
SELECT "test_query_table"."id",
       "test_query_table"."number",
       "relations_8106139955393"."id",
       "relations_8106139955393"."name",
       "relations_8106139955393"."value",
       "relations_8106139955393"."querytable"
FROM "test_query_table"
         LEFT JOIN "test_relationship" AS "relations_8106139955393"
                   ON ("relations_8106139955393"."querytable" = "test_query_table"."id")
WHERE ("test_query_table"."id" IN (SELECT "test_query_table"."id"
                                   FROM "test_query_table"
                                   WHERE ("test_query_table"."id" > 0)
                                   ORDER BY "test_query_table"."id"
                                   LIMIT 3 OFFSET 0))
ORDER BY "test_query_table"."number" DESC;
"""


def _setup_data():
    TestQueryTable.truncate()
    first = TestQueryTable.insert(number=0)
    second = TestQueryTable.insert(number=1)
    TestQueryTable.insert(number=2)
    TestQueryTable.insert(number=3)
    TestQueryTable.insert(number=4)

    TestRelationship.insert(name="First Relation", querytable=first, value=3)
    TestRelationship.insert(name="Second Relation", querytable=first, value=3)
    TestRelationship.insert(name="Third Relation", querytable=first, value=3)
    TestRelationship.insert(name="Fourth Relation", querytable=first, value=3)

    TestRelationship.insert(name="First Relation", querytable=second, value=33)
    TestRelationship.insert(name="Second Relation", querytable=second, value=33)
    TestRelationship.insert(name="Third Relation", querytable=second, value=33)
    TestRelationship.insert(name="Fourth Relation", querytable=second, value=33)

    db.commit()


def test_where_builder():
    _setup_data()

    assert TestQueryTable.first().id == TestQueryTable.select().first().id

    builder = TestQueryTable.where(lambda row: row.number < 3).where(TestQueryTable.number > 1)
    results = builder.collect()
    assert len(results) == 1

    sql = builder._first()
    assert "LIMIT 1" in sql

    instance = results.first()

    assert instance.number == 2
    assert instance.other == "Something"

    assert isinstance(instance, TestQueryTable)

    results = builder.select(TestQueryTable.id, TestQueryTable.number).first()

    assert "number" in TestQueryTable
    assert "number" in results.__dict__
    assert "number" in results
    assert results.number == results["number"] == 2

    assert "other" in TestQueryTable
    assert "other" not in results.__dict__
    assert "other" not in results
    assert not results.other

    with pytest.raises(KeyError):
        assert not results["other"]

    assert not results.yet_another

    # delete all numbers above 2
    builder = TestQueryTable.where(lambda row: row.number > 2)
    assert builder._delete()
    assert isinstance(builder._delete(), str)

    result = builder.delete()
    assert len(result) == 2  # -> 3 and 4

    assert TestQueryTable.where(lambda row: row.number > 99).delete() == []  # nothing deleted

    assert TestQueryTable.count() == 3 == len(TestQueryTable.collect())  # 0 - 2

    builder = TestQueryTable.where(lambda row: row.number == 0)
    sql = builder._update(number=5)
    assert sql
    assert isinstance(sql, str)
    assert "number" in sql
    assert "5" in sql

    result = builder.update(number=5)
    assert TestQueryTable.where(lambda row: row.number == -1).update(number=5) == []  # nothing updated
    assert result == [1]  # id 1 updated

    assert TestQueryTable(1).number == 5

    success = TestQueryTable.select().collect_or_fail()
    assert len(success) == 3

    # test OR

    assert TestQueryTable.where(lambda row: row.id == 1, lambda row: row.id == 2).count() == 2
    assert TestQueryTable.where(lambda row: row.id == 1, lambda row: row.id == 99).count() == 1

    assert TestQueryTable.where(id=-1).first() is None
    with pytest.raises(ValueError):
        TestQueryTable.where(id=-1).collect_or_fail()

    with pytest.raises(TypeError):
        TestQueryTable.where(id=-1).collect_or_fail(TypeError("pytest"))

    with pytest.raises(ValueError):
        assert not TestQueryTable.where(id=-1).first_or_fail()

    with pytest.raises(TypeError):
        assert not TestQueryTable.where(id=-1).first_or_fail(TypeError("pytest"))

    # try to break stuff:

    with pytest.raises(ValueError):
        # illegal query:
        builder.where(Exception())

    with pytest.raises(EnvironmentError):
        # can't collect before defining on db!
        Undefined.collect()


def test_select():
    _setup_data()

    # all:
    full = TestQueryTable.where(lambda row: row.number > 0).join().select().first_or_fail()

    assert full.number
    assert full.other
    assert full.yet_another

    assert full.relations[0].name
    assert full.relations[0].value

    # specific fields:
    partial = (
        TestQueryTable.where(lambda row: row.number > 0)
        .join()
        .select(TestQueryTable.other, TestRelationship.value)
        .first_or_fail()
    )

    assert partial.other
    assert not partial.number
    assert not partial.yet_another

    assert not partial.relations[0].name
    assert partial.relations[0].value


def test_paginate():
    _setup_data()

    result = TestQueryTable.join(method="left").paginate(limit=1, page=1)

    result_two = result.next().previous()

    assert len(result) == 1 == len(result_two)
    assert len(result.first().relations) == 4 == len(result_two.first().relations)

    meta = result.metadata["pagination"]
    assert meta == result_two.metadata["pagination"]

    assert meta["current_page"] == 1
    assert meta["limit"] == 1
    assert meta["min_max"] == (0, 1)

    next_page = result.next()
    result = TestQueryTable.join(method="left").paginate(limit=1, page=2)

    assert len(result) == 1 == len(next_page)
    assert len(result.first().relations) == 4 == len(next_page.first().relations)

    meta = result.metadata["pagination"]
    assert meta == next_page.metadata["pagination"]

    assert meta["current_page"] == 2
    assert meta["limit"] == 1
    assert meta["min_max"] == (1, 2)

    next_page = result.next()

    sql = TestQueryTable.join(method="left")._paginate(limit=1, page=3)

    assert "LIMIT 1" in sql
    assert "OFFSET 2" in sql

    result = TestQueryTable.join(method="left").paginate(limit=1, page=3)

    assert len(result) == 1 == len(next_page) == len(result.data)
    assert len(result.first().relations) == 0 == len(next_page.first().relations)

    meta = result.metadata["pagination"]
    assert meta == next_page.metadata["pagination"]

    assert meta["current_page"] == 3
    assert meta["limit"] == 1
    assert meta["min_max"] == (2, 3)

    # final page (all items on 1 page is no prev or next):
    result = TestQueryTable.paginate(limit=10, page=1)

    with pytest.raises(StopIteration):
        result.previous()

    with pytest.raises(StopIteration):
        result.next()

    page_dict = result.as_dict()["pagination"]

    assert page_dict["has_next_page"] is False
    assert page_dict["has_prev_page"] is False
    assert page_dict["next_page"] is None
    assert page_dict["prev_page"] is None

    # what if no limit?
    all_rows = TestQueryTable.join(method="left").paginate(limit=0, page=1)
    assert len(all_rows) == 5
    assert all_rows.pagination["total_items"] == 5
    assert all_rows.pagination["total_pages"] == 1
    assert not all_rows.pagination["has_next_page"]


def test_chunking():
    _setup_data()

    total = 0
    size = 3
    for rows in TestQueryTable.chunk(size):
        assert rows
        assert len(rows) <= size
        total += len(rows)

    assert total == TestQueryTable.count()


def test_complex_join():
    _setup_data()

    # only rows that ahve a TestQueryTable with number == 1
    builder = TestRelationship.join(
        "test_query_table",
        method="inner",
        condition=(TestRelationship.querytable == TestQueryTable.id) & (TestQueryTable.number == 1),
    )

    assert builder.count() == 4 == len(builder.collect())

    assert builder.collect()

    # notation 2:

    builder = TestRelationship.join(
        TestQueryTable,
        method="inner",
        condition=lambda relation, query: (relation.querytable == query.id) & (query.number == 1),
    )

    assert builder.count() == 4 == len(builder.collect())

    sql = builder._collect()
    assert "JOIN" in sql
    assert "LEFT" not in sql
    assert "CROSS" not in sql

    rows = builder.collect()

    # value = 3 only ahppens for querytable.number == 0
    assert len(rows) == 4
    assert set(rows.column("value")) == {33}

    for row in rows:
        # note: other column name since it's a manual relationship:
        assert row.test_query_table.number == 1

    page = builder.paginate(limit=2)

    assert set(page.column("value")) == {33}
    assert len(page) == 2

    page = page.next()
    assert len(page) == 2
    assert set(page.column("value")) == {33}

    with pytest.raises(StopIteration):
        assert not page.next()

    # custom .on:

    builder = TestRelationship.join(
        TestQueryTable,
        on=lambda relation, query: query.on((relation.querytable == query.id) & (query.number == 1)),
    )

    sql = builder._collect()
    assert "JOIN" in sql
    assert "LEFT" in sql
    assert "INNER" not in sql
    assert "CROSS" not in sql

    assert builder.count() == 8 == len(builder.collect())  # all but with less extra's

    # notation 2:

    builder = TestRelationship.join(
        TestQueryTable,
        on=TestQueryTable.on((TestRelationship.querytable == TestQueryTable.id) & (TestQueryTable.number == 1)),
    )

    assert builder.count() == 8 == len(builder.collect())  # all but with less extra's

    for row in builder.collect():
        if row.test_query_table:
            assert row.test_query_table.number == 1

    # value errors:

    with pytest.raises(ValueError):
        TestRelationship.join(
            TestQueryTable,
            TestRelationship,
            condition=lambda relation, query: (relation.querytable == query.id) & (query.number == 1),
        )

    with pytest.raises(ValueError):
        TestRelationship.join(
            TestQueryTable,
            TestRelationship,
            method="inner",
            on=lambda relation, query: (relation.querytable == query.id) & (query.number == 1),
        )

    with pytest.raises(ValueError):
        TestRelationship.join(
            TestQueryTable,
            method="inner",
            condition=lambda relation, query: (relation.querytable == query.id) & (query.number == 1),
            on=lambda relation, query: (relation.querytable == query.id) & (query.number == 1),
        )


def test_reprs_and_bool():
    _setup_data()

    # NOTE: This logic changed: any 'empty' query table is False, and adding any conditions make it True.
    empty = TestQueryTable.where()
    notempty = TestQueryTable.where(id=1)
    assert (empty or notempty) is notempty
    assert (notempty or empty) is notempty
    assert not empty
    assert notempty
    assert TestQueryTable.where(id=10000000000)

    assert repr(TestQueryTable.where(id=1))
    assert str(TestQueryTable.where(id=1))


def test_orderby():
    _setup_data()

    base_qt = TestQueryTable.select(TestQueryTable.id, TestQueryTable.number)

    assert base_qt.count() == 5

    rows1 = base_qt.select(orderby=TestQueryTable.id).paginate(limit=3, page=1)
    rows2 = base_qt.select(orderby=TestQueryTable.id, limitby=(0, 3)).collect()
    assert [_.id for _ in rows1] == [_.id for _ in rows2] == [1, 2, 3]

    rows1 = base_qt.select(orderby=TestQueryTable.number).paginate(limit=3, page=1)
    rows2 = base_qt.select(orderby=TestQueryTable.number, limitby=(0, 3)).collect()
    assert [_.number for _ in rows1] == [_.number for _ in rows2] == [0, 1, 2]

    rows1 = base_qt.select(orderby=~TestQueryTable.number).paginate(limit=3, page=1)
    rows2 = base_qt.select(orderby=~TestQueryTable.number, limitby=(0, 3)).collect()
    assert [_.number for _ in rows1] == [_.number for _ in rows2] == [4, 3, 2]

    joined_qt = base_qt.join()

    assert joined_qt.count() == 5  # left join shouldn't change count

    rows1 = joined_qt.select(orderby=TestQueryTable.id).paginate(limit=3, page=1)
    rows2 = joined_qt.select(orderby=TestQueryTable.id, limitby=(0, 3)).collect()
    assert [_.id for _ in rows1] == [_.id for _ in rows2] == [1, 2, 3]

    rows1 = joined_qt.select(orderby=TestQueryTable.number).paginate(limit=3, page=1)
    rows2 = joined_qt.select(orderby=TestQueryTable.number, limitby=(0, 3)).collect()
    assert [_.number for _ in rows1] == [_.number for _ in rows2] == [0, 1, 2]

    rows1 = joined_qt.select(orderby=~TestQueryTable.number).paginate(limit=3, page=1)
    rows2 = joined_qt.select(orderby=~TestQueryTable.number, limitby=(0, 3)).collect()
    assert [_.number for _ in rows1] == [_.number for _ in rows2] == [4, 3, 2]

    rows1 = joined_qt.select(orderby=~TestQueryTable.number).paginate(limit=100, page=1)
    rows2 = joined_qt.select(orderby=~TestQueryTable.number, limitby=(0, 100)).collect()
    assert [_.number for _ in rows1] == [_.number for _ in rows2] == [4, 3, 2, 1, 0]


def test_execute():
    _setup_data()

    raw_execute = TestRelationship.select(
        TestRelationship.querytable.with_alias("query_table"),
        TestRelationship.querytable.count().with_alias("count"),
        groupby=TestRelationship.querytable,
    ).execute()  # ! no proper typing

    # 2 x 4:

    assert len(raw_execute) == 2

    for row in raw_execute:
        assert row["count"] == 4


def test_column():
    _setup_data()

    rows = TestRelationship.where(TestRelationship.value > 30).column(TestRelationship.value)

    assert len(rows) == 4
    assert set(rows) == {33}


def test_collect_with_extra_fields():
    _setup_data()
    builder = TestRelationship.select(TestRelationship.id, TestRelationship.name, TestRelationship.querytable.count())

    assert builder.execute()

    row = builder.first_or_fail()

    assert row.id
    assert row.name
    assert row._extra
    assert row[TestRelationship.querytable.count()]



import datetime as dt
import os
import shutil
import tempfile
import uuid
from pathlib import Path

import pytest

# from contextlib import chdir
from contextlib_chdir import chdir
from pydal2sql import generate_sql
from testcontainers.postgres import PostgresContainer

from src.typedal import TypeDAL, TypedField, TypedTable
from src.typedal.config import (
    _load_dotenv,
    _load_toml,
    expand_env_vars_into_toml_values,
    load_config,
)
from src.typedal.fields import PointField, TimestampField, UUIDField

postgres = PostgresContainer(
    dbname="postgres",
    username="someuser",
    password="somepass",
)


@pytest.fixture(scope="module", autouse=True)
def psql(request):
    postgres.ports = {
        5432: 9631,  # as set in valid.env
    }

    request.addfinalizer(postgres.stop)
    postgres.start()


@pytest.fixture
def at_temp_dir():
    with tempfile.TemporaryDirectory() as d:
        with chdir(d):
            yield d


def _load_db_after_setup(dialect: str):
    config = load_config()
    db = TypeDAL(attempts=1)
    assert db._uri == config.database

    assert f"'dialect': '{dialect}'" in repr(config)

    return True


def test_load_toml(at_temp_dir):
    base = Path("pyproject.toml")
    base.write_text("# empty")

    with pytest.warns(UserWarning):
        # it should warn because the toml is empty
        assert _load_toml(False) == ("", {})
        assert _load_toml(None) == (str(base.resolve().absolute()), {})
        assert _load_toml(str(base)) == ("pyproject.toml", {})
        assert _load_toml(".") == (str(base.resolve().absolute()), {})


def test_load_dotenv(at_temp_dir):
    base = Path(".env")
    base.write_text("# empty")

    assert _load_dotenv(False)[0] == ""
    assert _load_dotenv(None)[0] == str(base.resolve().absolute())
    assert _load_dotenv(str(base))[0] == ".env"
    assert _load_dotenv(".")[0] == ".env"


def test_load_empty_config(at_temp_dir):
    assert _load_db_after_setup("sqlite")


def test_load_toml_config(at_temp_dir):
    examples = Path(__file__).parent / "configs"
    shutil.copy(examples / "valid.toml", "./pyproject.toml")

    assert _load_db_after_setup("sqlite")


def test_load_env_config(at_temp_dir):
    examples = Path(__file__).parent / "configs"
    shutil.copy(examples / "valid.env", "./.env")

    assert _load_db_after_setup("postgres")


def test_load_simple_config(at_temp_dir):
    examples = Path(__file__).parent / "configs"
    shutil.copy(examples / "valid.env", "./.env")
    shutil.copy(examples / "simple.toml", "./pyproject.toml")

    assert _load_db_after_setup("postgres")


def test_load_both_config(at_temp_dir):
    examples = Path(__file__).parent / "configs"
    shutil.copy(examples / "valid.env", "./.env")
    shutil.copy(examples / "valid.toml", "./pyproject.toml")

    assert _load_db_after_setup("postgres")


def test_converting(at_temp_dir):
    from edwh_migrate import Config as MigrateConfig
    from pydal2sql.typer_support import Config as P2SConfig

    config = load_config()

    assert isinstance(config.to_migrate(), MigrateConfig)
    assert isinstance(config.to_pydal2sql(), P2SConfig)


def test_environ(at_temp_dir):
    os.environ["DB_URI"] = "sqlite:///tmp/db.sqlite"
    config = load_config(False, True)

    assert config.database == "sqlite:///tmp/db.sqlite"


def test_expand_env_vars():
    # str
    input_str = "${MYVALUE:default}"
    data = {"myvar": input_str}
    expand_env_vars_into_toml_values(data, {})
    assert data["myvar"] == input_str

    expand_env_vars_into_toml_values(data, {"unrelated": "data"})
    assert data["myvar"] == "default"

    data = {"myvar": input_str}
    expand_env_vars_into_toml_values(data, {"myvalue": "123"})

    assert data["myvar"] == "123"

    # list
    data = {"myvar": [input_str, input_str]}
    expand_env_vars_into_toml_values(data, {"myvalue": "456"})

    assert data["myvar"] == ["456", "456"]

    # dict
    data = {"myvar": {"value": input_str}}
    expand_env_vars_into_toml_values(data, {"myvalue": "789"})
    assert data["myvar"]["value"] == "789"

    # other - non-str
    data = {"myvar": None, "mynumber": 123}
    expand_env_vars_into_toml_values(data, {"myvalue": "789"})
    assert data["myvar"] is None
    assert data["mynumber"] == 123


# note: these are not really 'config' specific but we already have access to postgres here so good enough:


def test_timestamp_fields_sqlite(at_temp_dir):
    db = TypeDAL("sqlite:memory")

    class Timestamp(TypedTable):
        ts = TimestampField(default=dt.datetime.now)
        dt = TypedField(dt.datetime, default=dt.datetime.now)

    db.define(Timestamp)

    row = Timestamp.insert()

    # old:
    assert isinstance(row.dt, dt.datetime), "not a datetime"
    assert "." not in str(row.dt)  # no ms precision

    # new:
    assert isinstance(row.ts, dt.datetime), "not a datetime"
    assert "." in str(row.ts)  # ms precision

    assert '"ts" timestamp NOT NULL' in Timestamp._sql()


def test_timestamp_fields_psql(at_temp_dir):
    examples = Path(__file__).parent / "configs"
    shutil.copy(examples / "valid.env", "./.env")
    shutil.copy(examples / "simple.toml", "./pyproject.toml")

    assert _load_db_after_setup("postgres")
    db = TypeDAL(attempts=1)

    class Timestamp(TypedTable):
        ts = TimestampField(default=dt.datetime.now)
        dt = TypedField(dt.datetime, default=dt.datetime.now)

    db.define(Timestamp)

    row = Timestamp.insert()

    # old:
    assert isinstance(row.dt, dt.datetime), "not a datetime"
    assert "." not in str(row.dt)  # no ms precision

    # new:
    assert isinstance(row.ts, dt.datetime), "not a datetime"
    assert "." in str(row.ts)  # ms precision

    assert '"ts" timestamp NOT NULL' in Timestamp._sql()


def test_point_fields_sqlite(at_temp_dir):
    db = TypeDAL("sqlite:memory")

    class Point(TypedTable):
        pt = PointField()

    db.define(Point)

    row1 = Point.insert(pt=(1, 0))
    row2 = Point.insert(pt="(1, 0)")

    assert row1.pt == row2.pt

    x, y = row1.pt
    assert x == 1
    assert y == 0

    with pytest.raises(Exception):
        Point.insert(pt=123)

    with pytest.raises(Exception):
        Point.insert(pt="123")

    # note: psql will check this whereas sqlite won't:
    Point.insert(pt=())

    assert '"pt" point NOT NULL' in Point._sql()


def test_point_fields_psql(at_temp_dir):
    examples = Path(__file__).parent / "configs"
    shutil.copy(examples / "valid.env", "./.env")
    shutil.copy(examples / "simple.toml", "./pyproject.toml")

    assert _load_db_after_setup("postgres")
    db = TypeDAL(attempts=1)

    class Point(TypedTable):
        pt = PointField()

    db.define(Point)

    row1 = Point.insert(pt=(1, 0))
    row2 = Point.insert(pt="(1, 0)")

    assert row1.pt == row2.pt

    x, y = row1.pt
    assert x == 1
    assert y == 0

    with pytest.raises(Exception):
        Point.insert(pt=123)

    with pytest.raises(Exception):
        Point.insert(pt="123")

    # note: psql will check this whereas sqlite won't:

    with pytest.raises(Exception):
        Point.insert(pt=())

    assert '"pt" point NOT NULL' in Point._sql()


def test_uuid_fields_psql(at_temp_dir):
    examples = Path(__file__).parent / "configs"
    shutil.copy(examples / "valid.env", "./.env")
    shutil.copy(examples / "simple.toml", "./pyproject.toml")

    assert _load_db_after_setup("postgres")
    db = TypeDAL(attempts=1)

    class UUIDTable(TypedTable):
        gid = UUIDField(default=uuid.uuid4)

    db.define(UUIDTable)

    row = UUIDTable.insert()

    assert isinstance(row.gid, uuid.UUID)

    with pytest.raises(Exception):
        UUIDTable.insert(gid="not-a-uuid")

    assert '"gid" uuid NOT NULL' in UUIDTable._sql()


def test_uuid_fields_sqlite(at_temp_dir):
    db = TypeDAL("sqlite:memory")

    class UUIDTable(TypedTable):
        gid = UUIDField(default=uuid.uuid4)

    db.define(UUIDTable)

    row = UUIDTable.insert()

    assert isinstance(row.gid, uuid.UUID)

    with pytest.raises(Exception):
        UUIDTable.insert(gid="not-a-uuid")

    assert '"gid" uuid NOT NULL' in UUIDTable._sql()



"""
This file contains available Field types.
"""

import ast
import datetime as dt
import decimal
import typing
import uuid

from pydal.helpers.classes import SQLCustomType
from pydal.objects import Table
from typing_extensions import Unpack

from .core import TypeDAL, TypedField, TypedTable
from .types import FieldSettings

T = typing.TypeVar("T", bound=typing.Any)


## general


## specific
def StringField(**kw: Unpack[FieldSettings]) -> TypedField[str]:
    """
    Pydal type is string, Python type is str.
    """
    kw["type"] = "string"
    return TypedField(str, **kw)


String = StringField


def TextField(**kw: Unpack[FieldSettings]) -> TypedField[str]:
    """
    Pydal type is text, Python type is str.
    """
    kw["type"] = "text"
    return TypedField(str, **kw)


Text = TextField


def BlobField(**kw: Unpack[FieldSettings]) -> TypedField[bytes]:
    """
    Pydal type is blob, Python type is bytes.
    """
    kw["type"] = "blob"
    return TypedField(bytes, **kw)


Blob = BlobField


def BooleanField(**kw: Unpack[FieldSettings]) -> TypedField[bool]:
    """
    Pydal type is boolean, Python type is bool.
    """
    kw["type"] = "boolean"
    return TypedField(bool, **kw)


Boolean = BooleanField


def IntegerField(**kw: Unpack[FieldSettings]) -> TypedField[int]:
    """
    Pydal type is integer, Python type is int.
    """
    kw["type"] = "integer"
    return TypedField(int, **kw)


Integer = IntegerField


def DoubleField(**kw: Unpack[FieldSettings]) -> TypedField[float]:
    """
    Pydal type is double, Python type is float.
    """
    kw["type"] = "double"
    return TypedField(float, **kw)


Double = DoubleField


def DecimalField(n: int, m: int, **kw: Unpack[FieldSettings]) -> TypedField[decimal.Decimal]:
    """
    Pydal type is decimal, Python type is Decimal.
    """
    kw["type"] = f"decimal({n}, {m})"
    return TypedField(decimal.Decimal, **kw)


Decimal = DecimalField


def DateField(**kw: Unpack[FieldSettings]) -> TypedField[dt.date]:
    """
    Pydal type is date, Python type is datetime.date.
    """
    kw["type"] = "date"
    return TypedField(dt.date, **kw)


Date = DateField


def TimeField(**kw: Unpack[FieldSettings]) -> TypedField[dt.time]:
    """
    Pydal type is time, Python type is datetime.time.
    """
    kw["type"] = "time"
    return TypedField(dt.time, **kw)


Time = TimeField


def DatetimeField(**kw: Unpack[FieldSettings]) -> TypedField[dt.datetime]:
    """
    Pydal type is datetime, Python type is datetime.datetime.
    """
    kw["type"] = "datetime"
    return TypedField(dt.datetime, **kw)


Datetime = DatetimeField


def PasswordField(**kw: Unpack[FieldSettings]) -> TypedField[str]:
    """
    Pydal type is password, Python type is str.
    """
    kw["type"] = "password"
    return TypedField(str, **kw)


Password = PasswordField


def UploadField(**kw: Unpack[FieldSettings]) -> TypedField[str]:
    """
    Pydal type is upload, Python type is str.
    """
    kw["type"] = "upload"
    return TypedField(str, **kw)


Upload = UploadField

T_subclass = typing.TypeVar("T_subclass", TypedTable, Table)


def ReferenceField(
    other_table: str | typing.Type[TypedTable] | TypedTable | Table | T_subclass, **kw: Unpack[FieldSettings]
) -> TypedField[int]:
    """
    Pydal type is reference, Python type is int (id).
    """
    if isinstance(other_table, str):
        kw["type"] = f"reference {other_table}"
    elif isinstance(other_table, Table):
        # db.table
        kw["type"] = f"reference {other_table._tablename}"
    elif isinstance(other_table, type) and issubclass(other_table, TypedTable):
        # SomeTable
        snakename = TypeDAL.to_snake(other_table.__name__)
        kw["type"] = f"reference {snakename}"
    else:
        raise ValueError(f"Don't know what to do with {type(other_table)}")

    return TypedField(int, **kw)


Reference = ReferenceField


def ListStringField(**kw: Unpack[FieldSettings]) -> TypedField[list[str]]:
    """
    Pydal type is list:string, Python type is list of str.
    """
    kw["type"] = "list:string"
    return TypedField(list[str], **kw)


ListString = ListStringField


def ListIntegerField(**kw: Unpack[FieldSettings]) -> TypedField[list[int]]:
    """
    Pydal type is list:integer, Python type is list of int.
    """
    kw["type"] = "list:integer"
    return TypedField(list[int], **kw)


ListInteger = ListIntegerField


def ListReferenceField(other_table: str, **kw: Unpack[FieldSettings]) -> TypedField[list[int]]:
    """
    Pydal type is list:reference, Python type is list of int (id).
    """
    kw["type"] = f"list:reference {other_table}"
    return TypedField(list[int], **kw)


ListReference = ListReferenceField


def JSONField(**kw: Unpack[FieldSettings]) -> TypedField[object]:
    """
    Pydal type is json, Python type is object (can be anything JSON-encodable).
    """
    kw["type"] = "json"
    return TypedField(object, **kw)


def BigintField(**kw: Unpack[FieldSettings]) -> TypedField[int]:
    """
    Pydal type is bigint, Python type is int.
    """
    kw["type"] = "bigint"
    return TypedField(int, **kw)


Bigint = BigintField

## Custom:

NativeTimestampField = SQLCustomType(
    type="datetime",
    native="timestamp",
    encoder=lambda x: f"'{x}'",  # extra quotes
    # decoder=lambda x: x, # already parsed into datetime
)


def TimestampField(**kw: Unpack[FieldSettings]) -> TypedField[dt.datetime]:
    """
    Database type is timestamp, Python type is datetime.

    Advantage over the regular datetime type is that
    a timestamp has millisecond precision (2024-10-11 20:18:24.505194)
    whereas a regular datetime only has precision up to the second (2024-10-11 20:18:24)
    """
    kw["type"] = NativeTimestampField
    return TypedField(
        dt.datetime,
        **kw,
    )


NativePointField = SQLCustomType(
    type="string",
    native="point",
    encoder=str,
    decoder=ast.literal_eval,
)


def PointField(**kw: Unpack[FieldSettings]) -> TypedField[tuple[float, float]]:
    """
    Database type is point, Python type is tuple[float, float].
    """
    kw["type"] = NativePointField
    return TypedField(tuple[float, float], **kw)


NativeUUIDField = SQLCustomType(
    type="string",
    native="uuid",
    encoder=str,
    decoder=uuid.UUID,
)


def UUIDField(**kw: Unpack[FieldSettings]) -> TypedField[uuid.UUID]:
    """
    Database type is uuid, Python type is UUID.
    """
    kw["type"] = NativeUUIDField
    return TypedField(uuid.UUID, **kw)



"""
Typer CLI for TypeDAL.
"""

import sys
import typing
import warnings
from pathlib import Path
from typing import Optional

import tomli
from configuraptor import asdict
from configuraptor.alias import is_alias
from configuraptor.helpers import is_optional

from .helpers import match_strings
from .types import AnyDict

try:
    import edwh_migrate
    import pydal2sql  # noqa: F401
    import questionary
    import rich
    import tomlkit
    import typer
    from tabulate import tabulate
except ImportError as e:
    # ImportWarning is hidden by default
    warnings.warn(
        "`migrations` extra not installed. Please run `pip install typedal[migrations]` to fix this.",
        source=e,
        category=RuntimeWarning,
    )
    exit(127)  # command not found

from pydal2sql.typer_support import IS_DEBUG, with_exit_code
from pydal2sql.types import (
    DBType_Option,
    OptionalArgument,
    OutputFormat_Option,
    Tables_Option,
)
from pydal2sql_core import core_alter, core_create, core_stub
from typing_extensions import Never

from . import caching
from .__about__ import __version__
from .config import TypeDALConfig, _fill_defaults, load_config, transform
from .core import TypeDAL

app = typer.Typer(
    no_args_is_help=True,
)

questionary_types: dict[typing.Hashable, Optional[AnyDict]] = {
    str: {
        "type": "text",
        "validate": lambda text: True if len(text) > 0 else "Please enter a value",
    },
    Optional[str]: {
        "type": "text",
        # no validate because it's optional
    },
    bool: {
        "type": "confirm",
    },
    int: {"type": "text", "validate": lambda text: True if text.isdigit() else "Please enter a number"},
    # specific props:
    "dialect": {
        "type": "select",
        "choices": ["sqlite", "postgres", "mysql"],
    },
    "folder": {
        "type": "path",
        "message": "Database directory:",
        "only_directories": True,
        # "default": "",
    },
    "input": {
        "type": "path",
        "message": "Python file containing table definitions.",
        "file_filter": lambda file: "." not in file or file.endswith(".py"),
    },
    "output": {
        "type": "path",
        "message": "Python file where migrations will be written to.",
        "file_filter": lambda file: "." not in file or file.endswith(".py"),
    },
    # disabled props:
    "pyproject": None,  # internal
    "noop": None,  # only for debugging
    "connection": None,  # internal
    "migrate": None,  # will probably conflict
    "fake_migrate": None,  # only enable via config if required
}

T = typing.TypeVar("T")

notfound = object()


def _get_question(prop: str, annotation: typing.Type[T]) -> Optional[AnyDict]:  # pragma: no cover
    question = questionary_types.get(prop, notfound)
    if question is notfound:
        # None means skip the question, notfound means use the type default!
        question = questionary_types.get(annotation)  # type: ignore

    if not question:
        return None
    # make a copy so the original is not overwritten:
    return question.copy()  # type: ignore


def get_question(prop: str, annotation: typing.Type[T], default: T | None) -> Optional[T]:  # pragma: no cover
    """
    Generate a question based on a config property and prompt the user for it.
    """
    if not (question := _get_question(prop, annotation)):
        return default

    question["name"] = prop
    question["message"] = question.get("message", f"{prop}? ")
    default = typing.cast(T, default or question.get("default") or "")

    if annotation is int:
        default = typing.cast(T, str(default))

    response = questionary.unsafe_prompt([question], default=default)[prop]
    return typing.cast(T, response)


@app.command()
@with_exit_code(hide_tb=IS_DEBUG)
def setup(
    config_file: typing.Annotated[Optional[str], typer.Option("--config", "-c")] = None,
    minimal: bool = False,
) -> None:  # pragma: no cover
    """
    Setup a [tool.typedal] entry in the local pyproject.toml.
    """
    # 1. check if [tool.typedal] in pyproject.toml and ask missing questions (excl .env vars)
    # 2. else if [tool.migrate] and/or [tool.pydal2sql] exist in the config, ask the user with copied defaults
    # 3. else: ask the user every question or minimal questions based on cli arg

    config = load_config(config_file)

    toml_path = Path(config.pyproject)

    if not (config.pyproject and toml_path.exists()):
        # no pyproject.toml found!
        toml_path = toml_path if config.pyproject else Path("pyproject.toml")
        rich.print(f"[blue]Config toml doesn't exist yet, creating {toml_path}[/blue]", file=sys.stderr)
        toml_path.touch()

    toml_contents = toml_path.read_text()
    # tomli has native Python types, tomlkit doesn't but preserves comments
    toml_obj: AnyDict = tomli.loads(toml_contents)

    if "[tool.pydal2sql]" in toml_contents:
        mapping = {"": ""}  # <- placeholder

        extra_config = toml_obj["tool"]["pydal2sql"]
        extra_config = {mapping.get(k, k): v for k, v in extra_config.items()}
        extra_config.pop("format", None)  # always edwh-migrate
        config.update(**extra_config)

    if "[tool.migrate]" in toml_contents:
        mapping = {"migrate_uri": "database"}

        extra_config = toml_obj["tool"]["migrate"]
        extra_config = {mapping.get(k, k): v for k, v in extra_config.items()}

        config.update(**extra_config)

    if "[tool.typedal]" in toml_contents:
        section = toml_obj["tool"]["typedal"]
        config.update(**section, _overwrite=True)

    data = asdict(config, with_top_level_key=False)
    data["migrate"] = None  # determined based on existence of input/output file.

    for prop, annotation in TypeDALConfig.__annotations__.items():
        if is_alias(config.__class__, prop):
            # don't store aliases!
            data.pop(prop, None)
            continue

        if minimal and getattr(config, prop, None) not in (None, "") or is_optional(annotation):
            # property already present or not required, SKIP!
            data[prop] = getattr(config, prop, None)
            continue

        _fill_defaults(data, prop, data.get(prop))
        default_value = data.get(prop, None)
        answer: typing.Any = get_question(prop, annotation, default_value)

        if isinstance(answer, str):
            answer = answer.strip()

        if annotation is bool:
            answer = bool(answer)
        elif annotation is int:
            answer = int(answer)

        config.update(**{prop: answer})
        data[prop] = answer

    for prop in TypeDALConfig.__annotations__:
        transform(data, prop)

    with toml_path.open("r") as f:
        old_contents: AnyDict = tomlkit.load(f)

    if "tool" not in old_contents:
        old_contents["tool"] = {}

    data.pop("pyproject", None)
    data.pop("connection", None)

    # ignore any None:
    old_contents["tool"]["typedal"] = {k: v for k, v in data.items() if v is not None}

    with toml_path.open("w") as f:
        tomlkit.dump(old_contents, f)

    rich.print(f"[green]Wrote updated config to {toml_path}![/green]")


@app.command(name="migrations.generate")
@with_exit_code(hide_tb=IS_DEBUG)
def generate_migrations(
    connection: typing.Annotated[str, typer.Option("--connection", "-c")] = None,
    filename_before: OptionalArgument[str] = None,
    filename_after: OptionalArgument[str] = None,
    dialect: DBType_Option = None,
    tables: Tables_Option = None,
    magic: Optional[bool] = None,
    noop: Optional[bool] = None,
    function: Optional[str] = None,
    output_format: OutputFormat_Option = None,
    output_file: Optional[str] = None,
    dry_run: bool = False,
) -> bool:  # pragma: no cover
    """
    Run pydal2sql based on the typedal config.
    """
    # 1. choose CREATE or ALTER based on whether 'output' exists?
    # 2. pass right args based on 'config' to function chosen in 1.
    generic_config = load_config(connection)
    pydal2sql_config = generic_config.to_pydal2sql()
    pydal2sql_config.update(
        magic=magic,
        noop=noop,
        tables=tables,
        db_type=dialect.value if dialect else None,
        function=function,
        format=output_format,
        input=filename_before,
        output=output_file,
        _skip_none=True,
    )

    if pydal2sql_config.output and Path(pydal2sql_config.output).exists():
        if dry_run:
            print("Would run `pyda2sql alter` with config", asdict(pydal2sql_config), file=sys.stderr)
            sys.stderr.flush()

            return True
        else:  # pragma: no cover
            return core_alter(
                pydal2sql_config.input,
                filename_after or pydal2sql_config.input,
                db_type=pydal2sql_config.db_type,
                tables=pydal2sql_config.tables,
                noop=pydal2sql_config.noop,
                magic=pydal2sql_config.magic,
                function=pydal2sql_config.function,
                output_format=pydal2sql_config.format,
                output_file=pydal2sql_config.output,
            )
    else:
        if dry_run:
            print("Would run `pyda2sql create` with config", asdict(pydal2sql_config), file=sys.stderr)
            sys.stderr.flush()

            return True
        else:  # pragma: no cover
            return core_create(
                filename=pydal2sql_config.input,
                db_type=pydal2sql_config.db_type,
                tables=pydal2sql_config.tables,
                noop=pydal2sql_config.noop,
                magic=pydal2sql_config.magic,
                function=pydal2sql_config.function,
                output_format=pydal2sql_config.format,
                output_file=pydal2sql_config.output,
            )


@app.command(name="migrations.run")
@with_exit_code(hide_tb=IS_DEBUG)
def run_migrations(
    connection: typing.Annotated[str, typer.Option("--connection", "-c")] = None,
    migrations_file: OptionalArgument[str] = None,
    db_uri: Optional[str] = None,
    db_folder: Optional[str] = None,
    schema_version: Optional[str] = None,
    redis_host: Optional[str] = None,
    migrate_cat_command: Optional[str] = None,
    database_to_restore: Optional[str] = None,
    migrate_table: Optional[str] = None,
    flag_location: Optional[str] = None,
    schema: Optional[str] = None,
    create_flag_location: Optional[bool] = None,
    dry_run: bool = False,
) -> bool:  # pragma: no cover
    """
    Run edwh-migrate based on the typedal config.
    """
    # 1. build migrate Config from TypeDAL config
    # 2. import right file
    # 3. `activate_migrations`
    generic_config = load_config(connection)
    migrate_config = generic_config.to_migrate()

    migrate_config.update(
        migrate_uri=db_uri,
        schema_version=schema_version,
        redis_host=redis_host,
        migrate_cat_command=migrate_cat_command,
        database_to_restore=database_to_restore,
        migrate_table=migrate_table,
        flag_location=flag_location,
        schema=schema,
        create_flag_location=create_flag_location,
        db_folder=db_folder,
        migrations_file=migrations_file,
        _skip_none=True,
    )

    if dry_run:
        print("Would run `migrate` with config", asdict(migrate_config), file=sys.stderr)
    else:  # pragma: no cover
        edwh_migrate.console_hook([], config=migrate_config)
    return True


@app.command(name="migrations.fake")
@with_exit_code(hide_tb=IS_DEBUG)
def fake_migrations(
    names: typing.Annotated[list[str], typer.Argument()] = None,
    all: bool = False,  # noqa: A002
    connection: typing.Annotated[str, typer.Option("--connection", "-c")] = None,
    migrations_file: Optional[str] = None,
    db_uri: Optional[str] = None,
    db_folder: Optional[str] = None,
    migrate_table: Optional[str] = None,
    dry_run: bool = False,
) -> int:  # pragma: no cover
    """
    Mark one or more migrations as completed in the database, without executing the SQL code.

    glob is supported in 'names'
    """
    if not (names or all):
        rich.print("Please provide one or more migration names, or pass --all to fake all.")
        return 1

    generic_config = load_config(connection)
    migrate_config = generic_config.to_migrate()

    migrate_config.update(
        migrate_uri=db_uri,
        migrate_table=migrate_table,
        db_folder=db_folder,
        migrations_file=migrations_file,
        _skip_none=True,
    )

    migrations = edwh_migrate.list_migrations(migrate_config)

    migration_names = list(migrations.keys())

    to_fake = migration_names if all else match_strings(names or [], migration_names)

    try:
        db = edwh_migrate.setup_db(config=migrate_config)
    except edwh_migrate.migrate.DatabaseNotYetInitialized:
        db = edwh_migrate.setup_db(
            config=migrate_config, migrate=True, migrate_enabled=True, remove_migrate_tablefile=True
        )

    previously_migrated = (
        db(
            db.ewh_implemented_features.name.belongs(to_fake)
            & (db.ewh_implemented_features.installed == True)  # noqa E712
        )
        .select(db.ewh_implemented_features.name)
        .column("name")
    )

    if dry_run:
        rich.print("Would migrate these:", [_ for _ in to_fake if _ not in previously_migrated])
        return 0

    n = len(to_fake)
    print(f"{len(previously_migrated)} / {n} were already installed.")

    for name in to_fake:
        if name in previously_migrated:
            continue

        edwh_migrate.mark_migration(db, name=name, installed=True)

    db.commit()
    rich.print(f"Faked {n} new migrations.")
    return 0


@app.command(name="migrations.stub")
@with_exit_code(hide_tb=IS_DEBUG)
def migrations_stub(
    migration_name: typing.Annotated[str, typer.Argument()] = "stub_migration",
    connection: typing.Annotated[str, typer.Option("--connection", "-c")] = None,
    output_format: OutputFormat_Option = None,
    output_file: Optional[str] = None,
    dry_run: typing.Annotated[bool, typer.Option("--dry", "--dry-run")] = False,
    is_pydal: typing.Annotated[bool, typer.Option("--pydal", "-p")] = False,
    # defaults to is_typedal of course
) -> int:
    """
    Create an empty migration via pydal2sql.
    """
    generic_config = load_config(connection)
    pydal2sql_config = generic_config.to_pydal2sql()
    pydal2sql_config.update(
        format=output_format,
        output=output_file,
        _skip_none=True,
    )

    core_stub(
        migration_name,  # raw, without date or number
        output_format=pydal2sql_config.format,
        output_file=pydal2sql_config.output or None,
        dry_run=dry_run,
        is_typedal=not is_pydal,
    )
    return 0


AnyNestedDict: typing.TypeAlias = dict[str, AnyDict]


def tabulate_data(data: AnyNestedDict) -> None:
    """
    Print a nested dict of data in a nice, human-readable table.
    """
    flattened_data = []
    for key, inner_dict in data.items():
        temp_dict = {"": key}
        temp_dict.update(inner_dict)
        flattened_data.append(temp_dict)

    # Display the tabulated data from the transposed dictionary
    print(tabulate(flattened_data, headers="keys"))


FormatOptions: typing.TypeAlias = typing.Literal["plaintext", "json", "yaml", "toml"]


def get_output_format(fmt: FormatOptions) -> typing.Callable[[AnyNestedDict], None]:
    """
    This function takes a format option as input and \
        returns a function that can be used to output data in the specified format.
    """
    match fmt:
        case "plaintext":
            output = tabulate_data
        case "json":

            def output(_data: AnyDict | AnyNestedDict) -> None:
                import json

                print(json.dumps(_data, indent=2))

        case "yaml":

            def output(_data: AnyDict | AnyNestedDict) -> None:
                import yaml

                print(yaml.dump(_data))

        case "toml":

            def output(_data: AnyDict | AnyNestedDict) -> None:
                import tomli_w

                print(tomli_w.dumps(_data))

        case _:
            options = typing.get_args(FormatOptions)
            raise ValueError(f"Invalid format '{fmt}'. Please choose one of {options}.")

    return output


@app.command(name="cache.stats")
@with_exit_code(hide_tb=IS_DEBUG)
def cache_stats(
    identifier: typing.Annotated[str, typer.Argument()] = "",
    connection: typing.Annotated[str, typer.Option("--connection", "-c")] = None,
    fmt: typing.Annotated[
        str, typer.Option("--format", "--fmt", "-f", help="plaintext (default) or json")
    ] = "plaintext",
) -> None:  # pragma: no cover
    """
    Collect caching stats.

    Examples:
        typedal cache.stats
        typedal cache.stats user
        typedal cache.stats user.3
    """
    config = load_config(connection)
    db = TypeDAL(config=config, migrate=False, fake_migrate=False)

    output = get_output_format(typing.cast(FormatOptions, fmt))

    data: AnyDict
    parts = identifier.split(".")
    match parts:
        case [] | [""]:
            # generic stats
            data = caching.calculate_stats(db)  # type: ignore
        case [table]:
            # table stats
            data = caching.table_stats(db, table)  # type: ignore
        case [table, row_id]:
            # row stats
            data = caching.row_stats(db, table, row_id)  # type: ignore
        case _:
            raise ValueError("Please use the format `table` or `table.id` for this command.")

    output(data)

    # todo:
    #  - sort by most dependencies
    #  - sort by biggest data
    #  - include size for table_stats, row_stats
    #  - group by table


@app.command(name="cache.clear")
@with_exit_code(hide_tb=IS_DEBUG)
def cache_clear(
    connection: typing.Annotated[str, typer.Option("--connection", "-c")] = None,
    purge: typing.Annotated[bool, typer.Option("--all", "--purge", "-p")] = False,
) -> None:  # pragma: no cover
    """
    Clear (expired) items from the cache.

    Args:
        connection (optional): [tool.typedal.<connection>]
        purge (default: no): remove all items, not only expired
    """
    config = load_config(connection)
    db = TypeDAL(config=config, migrate=False, fake_migrate=False)

    if purge:
        caching.clear_cache()
        print("Emptied cache")
    else:
        n = caching.clear_expired()
        print(f"Removed {n} expired from cache")

    db.commit()


def version_callback() -> Never:
    """
    --version requested!
    """
    print(f"pydal2sql Version: {__version__}")

    raise typer.Exit(0)


def config_callback() -> Never:
    """
    --show-config requested.
    """
    config = load_config()

    print(repr(config))

    raise typer.Exit(0)


@app.callback(invoke_without_command=True)
def main(
    _: typer.Context,
    # stops the program:
    show_config: bool = False,
    version: bool = False,
) -> None:
    """
    This script can be used to generate the create or alter sql from pydal or typedal.
    """
    if show_config:
        config_callback()
    elif version:
        version_callback()
    # else: just continue


if __name__ == "__main__":  # pragma: no cover
    app()



"""
ONLY USE IN COMBINATION WITH PY4WEB!
"""

import typing

import threadsafevariable
from configuraptor.abs import AnyType
from py4web.core import ICECUBE
from py4web.core import Fixture as _Fixture
from pydal.base import MetaDAL, hashlib_md5

from .core import TypeDAL
from .types import AnyDict
from .web2py_py4web_shared import AuthUser


class Fixture(_Fixture):  # type: ignore
    """
    Make mypy happy.
    """


class PY4WEB_DAL_SINGLETON(MetaDAL):
    _instances: typing.ClassVar[typing.MutableMapping[str, AnyType]] = {}

    def __call__(cls, uri: str, *args: typing.Any, **kwargs: typing.Any) -> AnyType:
        db_uid = kwargs.get("db_uid", hashlib_md5(repr(uri)).hexdigest())
        if db_uid not in cls._instances:
            cls._instances[db_uid] = super().__call__(*args, **kwargs)

        return cls._instances[db_uid]

    def _clear(cls) -> None:
        cls._instances.clear()


class DAL(TypeDAL, Fixture, metaclass=PY4WEB_DAL_SINGLETON):  # pragma: no cover
    """
    Fixture similar to the py4web pydal fixture, but for typedal.
    """

    def on_request(self, _: AnyDict) -> None:
        """
        Make sure there is a database connection when a request comes in.
        """
        self.get_connection_from_pool_or_new()
        threadsafevariable.ThreadSafeVariable.restore(ICECUBE)

    def on_error(self, _: AnyDict) -> None:
        """
        Rollback db on error.
        """
        self.recycle_connection_in_pool_or_close("rollback")

    def on_success(self, _: AnyDict) -> None:
        """
        Commit db on success.
        """
        self.recycle_connection_in_pool_or_close("commit")


def setup_py4web_tables(db: TypeDAL) -> None:
    """
    Setup all the (default) required auth table.
    """
    db.define(AuthUser, migrate=False, redefine=True)


__all__ = [
    "AuthUser",
    "Fixture",
    "DAL",
    "setup_py4web_tables",
]



"""
Test (Typed)Table public API.
"""

import datetime as dt
import io
import json
from textwrap import dedent

import dateutil.parser
import pydal
import pytest
from pydal.objects import Expression

from src.typedal import TypeDAL, TypedField, TypedTable
from src.typedal.fields import IntegerField
from src.typedal.serializers import as_json

db = TypeDAL("sqlite:memory")


def test_both_styles_for_class():
    old_style = db.define_table(
        "old_style",
        pydal.Field("string_field"),
        pydal.Field("int_field", "integer"),
    )

    @db.define()
    class NewStyle(TypedTable):
        string_field: TypedField[str]
        int_field = IntegerField()

    assert old_style.as_dict()
    assert NewStyle.as_dict()

    assert NewStyle.as_dict().keys() == old_style.as_dict().keys()

    assert json.loads(NewStyle.as_json()).keys() == json.loads(old_style.as_json()).keys()

    assert as_json.encode(NewStyle) == NewStyle.as_json()

    # assert old_style.as_xml()
    # assert NewStyle.as_xml()

    assert old_style.as_yaml()
    assert NewStyle.as_yaml()

    bulk_data = [
        {"string_field": "String 1", "int_field": 1},
        {"string_field": "String 2", "int_field": 2},
    ]
    ids = old_style.bulk_insert(bulk_data)
    assert ids == [_.id for _ in NewStyle.bulk_insert(bulk_data)]

    assert old_style.create_index("first-index-oldstyle", "int_field")
    assert NewStyle.create_index("first-index-newstyle", "int_field")

    assert old_style.drop_index("first-index-oldstyle")
    assert NewStyle.drop_index("first-index-newstyle")

    assert old_style.drop_index("first-index-oldstyle", if_exists=True)
    assert NewStyle.drop_index("first-index-newstyle", if_exists=True)

    assert old_style.fields[1] == "string_field"
    assert NewStyle.fields[1] == "string_field"

    demo_csv = dedent(
        """\
    string_field,int_field
    field 3,3
    field 4,4
    field n,-1\
    """
    )

    old_style.import_from_csv_file(io.StringIO(demo_csv), validate=True)
    NewStyle.import_from_csv_file(io.StringIO(demo_csv), validate=True)

    assert db(old_style).count() == 5
    assert NewStyle.count() == 5

    assert old_style.insert(string_field="field 6", int_field=6) == 6
    instance = NewStyle.insert(string_field="field 6", int_field=6)
    assert instance.id == 6

    with pytest.raises(RuntimeError):
        old_style.update(old_style.int_field == 6, int_field=7)

    assert NewStyle.update(NewStyle.int_field == 6000, int_field=7) is None

    same_as_before = NewStyle.update(NewStyle.int_field == 6, int_field=7).update_record(int_field=6)

    assert same_as_before.int_field == 6

    assert isinstance(old_style.on(old_style.id == 1), Expression)
    assert isinstance(NewStyle.on(NewStyle.id == 1), Expression)

    assert old_style.query_name()[0] == '"old_style"' == old_style.sql_fullref
    assert NewStyle.query_name()[0] == '"new_style"' == NewStyle.sql_fullref

    assert not old_style.update_or_insert(old_style.id == 5, string_field="field 5", int_field=5)  # update yields None
    instance = NewStyle.update_or_insert(old_style.id == 5, string_field="field 5", int_field=5)

    assert instance.int_field == 5

    assert old_style.update_or_insert(old_style.id == 7, string_field="field 7", int_field=7)  # insert yields id
    instance = NewStyle.update_or_insert(old_style.id == 7, string_field="field 7", int_field=7)

    assert instance.id == 7
    assert instance.int_field == 7

    # query with dict:
    assert old_style.update_or_insert(dict(string_field="field 8", int_field=8), string_field="field 8", int_field=8)
    assert NewStyle.update_or_insert(dict(string_field="field 8", int_field=8), string_field="field 8", int_field=8)

    # without query:
    assert old_style.update_or_insert(string_field="field 9", int_field=9)
    assert NewStyle.update_or_insert(string_field="field 9", int_field=9)

    old_style.truncate()
    NewStyle.truncate()

    assert db(old_style).count() == 0
    assert NewStyle.count() == 0

    with pytest.raises(Exception):
        old_style.insert(string_field=123, int_field="abc")

    with pytest.raises(Exception):
        NewStyle.insert(string_field=123, int_field="abc")

    assert old_style.validate_and_insert(string_field=123, int_field="abc")["errors"]
    instance, errors = NewStyle.validate_and_insert(string_field=123, int_field="abc")
    assert not instance
    assert errors

    assert old_style.validate_and_insert(string_field="123", int_field=123)["id"]
    instance, errors = NewStyle.validate_and_insert(string_field="123", int_field=123)
    assert instance
    assert not errors

    assert old_style.validate_and_update(old_style.id == 1, string_field=123, int_field="abc")["errors"]
    assert old_style.validate_and_update(old_style.id == 1, string_field="123", int_field=123)["id"]

    # errors - int field is a string
    instance, errors = NewStyle.validate_and_update(NewStyle.id == 1, string_field=123, int_field="abc")
    assert not instance
    assert errors

    # errors - required field is None
    instance, errors = NewStyle.validate_and_update(NewStyle.id == 1, string_field=None, int_field=None)
    assert not instance
    assert errors

    # success - types match
    instance, errors = NewStyle.validate_and_update(NewStyle.id == 1, string_field="123", int_field=123)
    assert instance
    assert not errors

    # no instance because id 99 doesn't exist, but also no errors because types match:
    instance, errors = NewStyle.validate_and_update(NewStyle.id == 99, string_field="123", int_field=123)
    assert not instance
    assert not errors

    assert old_style.validate_and_update_or_insert(old_style.id == 1, string_field=123, int_field="abc")["errors"]
    assert old_style.validate_and_update_or_insert(old_style.id == 101, string_field=123, int_field="abc")["errors"]

    assert old_style.validate_and_update_or_insert(old_style.id == 1, string_field="123", int_field=123)["id"]
    assert old_style.validate_and_update_or_insert(old_style.id == 101, string_field="123", int_field=123)["id"]

    instance, errors = NewStyle.validate_and_update_or_insert(old_style.id == 1, string_field=123, int_field="abc")

    assert not instance
    assert errors

    instance, errors = NewStyle.validate_and_update_or_insert(old_style.id == 99, string_field=123, int_field="abc")

    assert not instance
    assert errors

    instance, errors = NewStyle.validate_and_update_or_insert(old_style.id == 1, string_field="123", int_field=123)
    assert instance
    assert not errors

    instance, errors = NewStyle.validate_and_update_or_insert(old_style.id == 101, string_field="123", int_field=123)

    assert instance
    assert not errors

    assert len(db(old_style).select()) == 2
    assert len(NewStyle.all()) == 2

    assert isinstance(old_style.with_alias("aliased_old_style"), pydal.objects.Table)
    assert isinstance(NewStyle.with_alias("aliased_old_style"), pydal.objects.Table)

    old_style.drop()
    NewStyle.drop()

    with pytest.raises(Exception):
        old_style.drop()

    with pytest.raises(Exception):
        NewStyle.drop()


def test_table_with_methods():
    simulated_today = dateutil.parser.parse("2020-01-01 00:00").date()

    @db.define()
    class TableWithMethods(TypedTable):
        birthday: dt.date

        @property
        def today(self):
            return dt.date.today()

        @staticmethod
        def get_age(dob: dt.date, today: dt.date) -> int:
            if today.month < dob.month or (today.month == dob.month and today.day < dob.day):
                return today.year - dob.year - 1
            else:
                return today.year - dob.year

        @classmethod
        def get_tablename(cls):
            return str(cls._table)

        def age(self, today: dt.date = None):
            # https://stackoverflow.com/questions/765797/convert-timedelta-to-years
            today = today or self.today
            dob = self.birthday

            return self.get_age(dob, today)

        def _as_dict(self, **kwargs):
            # custom as_dict behavior! Used by as_json()
            return super()._as_dict() | {"age": self.age(simulated_today)}

    row = TableWithMethods.insert(birthday="2000-01-01")

    assert row.age(simulated_today) == 20

    assert TableWithMethods.get_tablename() == "table_with_methods"

    # test custom JSON dumping behavior:
    assert row.as_dict()["age"] == 20

    loaded = json.loads(row.as_json())
    assert loaded["birthday"]
    assert loaded["age"] == 20

    # and from a level up:
    dumped = TableWithMethods.all().as_json()

    loaded = json.loads(dumped)
    assert loaded[0]["age"] == 20



import json
import tempfile
from contextlib import chdir

from pydal.validators import IS_EMAIL, IS_NOT_IN_DB

from src.typedal import TypedTable
from src.typedal.for_py4web import DAL, AuthUser, setup_py4web_tables
from src.typedal.serializers import as_json

db = DAL("sqlite:memory")


def test_P4W():
    assert hasattr(DAL, "on_request")
    assert hasattr(db, "on_request")


def test_serialize():
    @db.define()
    class JsonTable(TypedTable):
        field: str

    row = JsonTable.insert(field="Hey")

    l = json.loads(as_json.encode(JsonTable))
    assert l
    assert isinstance(l, dict)
    l = json.loads(as_json.encode(row))
    assert isinstance(l, dict)
    assert l

    l = json.loads(as_json.encode(JsonTable.paginate(limit=1)))

    assert l["pagination"]
    assert l["data"]


def test_auth_user():
    setup_py4web_tables(db)
    requirements = AuthUser.email.requires

    assert requirements
    assert isinstance(requirements, tuple)
    assert len(requirements) == 2

    assert isinstance(requirements[0], IS_EMAIL)
    assert isinstance(requirements[1], IS_NOT_IN_DB)


def test_py4web_dal_singleton():
    with tempfile.TemporaryDirectory() as d:
        with chdir(d):
            # note: typedal caching is disabled here because otherwise the cache table might send logs to {d}/log.sql
            # which can lead to problems in other tests :(
            db_1a = DAL("sqlite:memory", enable_typedal_caching=False)
            db_1b = DAL("sqlite:memory", enable_typedal_caching=False)

            db_2a = DAL("sqlite://test_py4web_dal_singleton", folder=d, enable_typedal_caching=False)
            db_2b = DAL("sqlite://test_py4web_dal_singleton", folder=d, enable_typedal_caching=False)

            assert db_1a is db_1b
            assert db_2a is db_2b
            assert db_1a is not db_2a
            assert db_1b is not db_2b

    # reset singletons for later use:
    DAL._clear()



import datetime

from pydal import DAL, Field

from typedal.helpers import utcnow

db = DAL("sqlite:memory")

### basic examples

db.define_table("person",
                Field("name", "string"),
                Field("age", "integer", default=18),
                Field("nicknames", "list:string")
                )

db.define_table("pet",
                Field("name", "string"),
                Field("owners", "list:reference person")
                )

db.person.insert(name="Henk", age=44, nicknames=["Henkie", "Gekke Henk"])
db.person.insert(name="Ingrid", age=47, nicknames=[])

henk = db(db.person.name == "Henk").select().first()
ingrid = db(db.person.name == "Ingrid").select().first()
print(henk, ingrid)

db.pet.insert(name="Max", owners=[henk, ingrid])

max = db.pet(name="Max")
print(max)

people = db(db.person.id > 0).select()

for person in people:
    print(person.name)

### example with all possible field types;
db.define_table("other_table")

db.define_table("all_fields",
                Field("string", "string"),
                Field("text", "text"),
                Field("blob", "blob"),
                Field("boolean", "boolean"),
                Field("integer", "integer"),
                Field("double", "double"),
                Field("decimal", "decimal(2,3)"),
                Field("date", "date"),
                Field("time", "time"),
                Field("datetime", "datetime"),
                Field("password", "password"),
                Field("upload", "upload", uploadfield="upload_data"),
                Field("upload_data", "blob"),
                Field("reference", "reference other_table"),
                Field("list_string", "list:string"),
                Field("list_integer", "list:integer"),
                Field("list_reference", "list:reference other_table"),
                Field("json", "json"),
                Field("bigint", "bigint"),
                # The big-id and, big-reference are only supported by some of the database engines and are experimental.
                )

now = utcnow()

db.other_table.insert()
db.other_table.insert()
other1 = db.other_table(id=1)
other2 = db.other_table(id=2)

with open('example_old.py', 'rb') as stream:
    db.all_fields.insert(
        string="hi",
        text="hi but longer",
        blob=b"\x23",
        boolean=True,
        integer=1,
        double=1.11111111111111111111111111111,
        decimal=1.11111111111111111111111111111,
        date=now.date(),
        time=now.time(),
        datetime=now,
        password="secret",
        upload=stream,
        upload_data=stream.read(),
        reference=other1,
        list_string=["hi", "there"],
        list_integer=[1, 2],
        list_reference=[other1, other2],
        json={'hi': 'there'},
        bigint=42,
    )

row = db.all_fields(string="hi")
for field in row:
    print(field, type(row[field]))
